/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.apilib.common.command.grpc;

import java.util.List;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandId;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.OzgCloudCommandStatus;
import de.ozgcloud.apilib.common.command.OzgCloudCreateSubCommandsRequest;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsRequest;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@RequiredArgsConstructor
public class GrpcOzgCloudCommandService implements OzgCloudCommandService {

	private static final int WAIT_TIME_MS = 500;
	public static final int DEFAULT_COMMAND_REQUEST_THRESHOLD_MILLIS = 10000;

	@GrpcClient("ozgcloud-command-manager")
	private final CommandServiceBlockingStub commandServiceStub;

	private final CommandMapper mapper;

	private final OzgCloudCallContextProvider contextProvider;

	private final int commandRequestThresholdMillis;

	@Override
	public OzgCloudCommand createAndWaitUntilDone(OzgCloudCommand commandToCreate) {
		return waitUntilDone(create(commandToCreate));
	}

	@Override
	public OzgCloudCommand create(OzgCloudCommand commandToCreate) {
		var created = getCommandServiceStub().createCommand(mapper.buildCreateCommandRequest(commandToCreate));
		return mapper.fromGrpc(created.getCommand());
	}

	OzgCloudCommand waitUntilDone(OzgCloudCommand commandToWaitFor) {
		var command = commandToWaitFor;
		var timeout = System.currentTimeMillis() + commandRequestThresholdMillis;
		while (!command.getStatus().isFinalState() && System.currentTimeMillis() < timeout) {
			synchronized (this) {
				try {
					wait(WAIT_TIME_MS);
					command = reloadCommand(command.getId());
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
				}
			}
		}
		verifyCommand(command);
		return command;
	}

	OzgCloudCommand reloadCommand(OzgCloudCommandId commandId) {
		GrpcGetCommandRequest request = GrpcGetCommandRequest.newBuilder().setId(commandId.toString()).build();
		return mapper.fromGrpc(getCommandServiceStub().getCommand(request));
	}

	void verifyCommand(OzgCloudCommand command) {
		if (OzgCloudCommandStatus.ERROR.equals(command.getStatus())) {
			throw new TechnicalException("Command (id=%s) failed: %s".formatted(command.getId(), command.getErrorMessage()));
		}
	}

	public List<OzgCloudCommand> addSubCommands(OzgCloudCreateSubCommandsRequest request) {
		var response = getCommandServiceStub().addSubCommands(mapper.toAddSubCommandsRequest(request));
		return response.getCommandList().stream().map(mapper::fromGrpc).toList();
	}

	public boolean existsPendingCommands(OzgCloudVorgangId vorgangId) {
		var response = getCommandServiceStub().existsPendingCommands(buildExistsPendingCommandRequest(vorgangId));
		return response.getExistsPendingCommands();
	}

	GrpcExistsPendingCommandsRequest buildExistsPendingCommandRequest(OzgCloudVorgangId vorgangId) {
		return GrpcExistsPendingCommandsRequest.newBuilder().setVorgangId(vorgangId.toString()).build();
	}

	public List<OzgCloudCommand> getPendingCommands(OzgCloudVorgangId vorgangId) {
		var response = getCommandServiceStub().getPendingCommands(buildGetPendingCommandRequest(vorgangId));
		return response.getCommandList().stream().map(mapper::fromGrpc).toList();
	}

	GrpcGetPendingCommandsRequest buildGetPendingCommandRequest(OzgCloudVorgangId vorgangId) {
		return GrpcGetPendingCommandsRequest.newBuilder().setVorgangId(vorgangId.toString()).build();
	}

	CommandServiceBlockingStub getCommandServiceStub() {
		return commandServiceStub.withInterceptors(new OzgCloudCallContextAttachingInterceptor(contextProvider));
	}

}
