/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.apilib.file.grpc;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.errorhandling.NotFoundException;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.apilib.file.OzgCloudUploadFileTestFactory;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class GrpcOzgCloudFileServiceTest {

	@Spy
	@InjectMocks
	private GrpcOzgCloudFileService service;

	@Mock
	private BinaryFileServiceBlockingStub blockingStub;
	@Mock
	private BinaryFileServiceStub asyncServiceStub;
	@Spy
	private OzgCloudFileMapper mapper = Mappers.getMapper(OzgCloudFileMapper.class);

	@Nested
	class TestGetFile {

		@Nested
		class CallGetFile {
			@BeforeEach
			void init() {
				when(blockingStub.withInterceptors(any())).thenReturn(blockingStub);
				when(blockingStub.findBinaryFilesMetaData(any())).thenReturn(
					GrpcFindFilesResponse.newBuilder().addFile(GrpcOzgFileTestFactory.create()).build());
			}

			@Test
			void shouldCallStub() {
				service.getFile(OzgCloudFileTestFactory.ID);

				verify(blockingStub).findBinaryFilesMetaData(any());
			}

			@Test
			void shouldCallMapper() {
				service.getFile(OzgCloudFileTestFactory.ID);

				verify(mapper).fromGrpc(any());
			}
		}

		@Nested
		class BuildFindFileRequest {

			@Test
			void shouldAddFileId() {
				var request = service.buildFindFileRequest(OzgCloudFileTestFactory.ID);

				assertThat(request.getFileIdList()).hasSize(1).first().isEqualTo(OzgCloudFileTestFactory.ID_STR);
			}
		}

		@Nested
		class GetSingleResponse {
			@Test
			void shouldReturnSingleFile() {
				var result = service.getSingleResponse(GrpcFindFilesResponseTestFactory.create(), OzgCloudFileTestFactory.ID);

				assertThat(result).usingRecursiveComparison().isEqualTo(GrpcOzgFileTestFactory.create());
			}

			@Test
			void shouldReturnFirstFile() {
				var response = GrpcFindFilesResponseTestFactory.createBuilder().addFile(GrpcOzgFileTestFactory.createBuilder().setId("2").build())
					.build();

				var result = service.getSingleResponse(response, OzgCloudFileTestFactory.ID);

				assertThat(result).usingRecursiveComparison().isEqualTo(GrpcOzgFileTestFactory.create());
			}

			@Test
			void shouldThrowNotFound() {
				var response = GrpcFindFilesResponseTestFactory.createBuilder().clearFile().build();

				assertThatThrownBy(() -> service.getSingleResponse(response, OzgCloudFileTestFactory.ID))
					.isInstanceOf(NotFoundException.class);
			}
		}
	}

	@Nested
	class TestUploadFile {

		@Nested
		class TestUploadFileData {

			@Mock
			private FileSender<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse> fileSender;
			@Mock
			private GrpcUploadBinaryFileRequest uploadBinaryFileRequest;

			GrpcUploadBinaryFileResponse response =
				GrpcUploadBinaryFileResponse.newBuilder().setFileId(OzgCloudFileTestFactory.ID_STR)
				.build();

			@BeforeEach
			void init() {
				doReturn(response).when(service).waitUntilFutureToComplete(any(), any());
			}

			@Test
			void shouldCallCreateSender() {
				when(fileSender.withMetaData(any())).thenReturn(fileSender);
				try (var uploadUtils = Mockito.mockStatic(GrpcFileUploadUtils.class)) {
					uploadUtils.when(() -> GrpcFileUploadUtils.createSender(any(), any(), any())).thenReturn(fileSender);
					var dataStream = InputStream.nullInputStream();

					service.uploadFile(OzgCloudUploadFileTestFactory.create(), dataStream);

					uploadUtils.verify(() -> GrpcFileUploadUtils.createSender(any(), eq(dataStream), any()));
				}
			}

			@Test
			void shouldCallBuildMetadataRequest() {
				when(asyncServiceStub.withInterceptors(any())).thenReturn(asyncServiceStub);
				var file = OzgCloudUploadFileTestFactory.create();

				service.uploadFile(file, InputStream.nullInputStream());

				verify(service).buildMetaDataRequest(file);
			}

			@Test
			void shoudlSetMetadata() {
				try (var uploadUtils = Mockito.mockStatic(GrpcFileUploadUtils.class)) {
					uploadUtils.when(() -> GrpcFileUploadUtils.createSender(any(), any(), any())).thenReturn(fileSender);
					when(fileSender.withMetaData(any())).thenReturn(fileSender);
					doReturn(uploadBinaryFileRequest).when(service).buildMetaDataRequest(any());

					service.uploadFile(OzgCloudUploadFileTestFactory.create(), InputStream.nullInputStream());

					verify(fileSender).withMetaData(uploadBinaryFileRequest);
				}
			}

			@Test
			void shouldCallSend() {
				try (var uploadUtils = Mockito.mockStatic(GrpcFileUploadUtils.class)) {
					uploadUtils.when(() -> GrpcFileUploadUtils.createSender(any(), any(), any())).thenReturn(fileSender);
					when(fileSender.withMetaData(any())).thenReturn(fileSender);

					service.uploadFile(OzgCloudUploadFileTestFactory.create(), InputStream.nullInputStream());

					verify(fileSender).send();
				}
			}

			@Test
			void shouldCallWaitUntilFutureToComplete() {
				try (var uploadUtils = Mockito.mockStatic(GrpcFileUploadUtils.class)) {
					uploadUtils.when(() -> GrpcFileUploadUtils.createSender(any(), any(), any())).thenReturn(fileSender);
					when(fileSender.withMetaData(any())).thenReturn(fileSender);
					when(fileSender.send()).thenReturn(fileSender);
					var dataStream = InputStream.nullInputStream();

					service.uploadFile(OzgCloudUploadFileTestFactory.create(), dataStream);

					verify(service).waitUntilFutureToComplete(fileSender, dataStream);
				}
			}

			@Test
			void shouldReturnResult() {
				when(asyncServiceStub.withInterceptors(any())).thenReturn(asyncServiceStub);

				var result = service.uploadFile(OzgCloudUploadFileTestFactory.create(), InputStream.nullInputStream());

				assertThat(result).isEqualTo(OzgCloudFileTestFactory.ID);
			}

			private GrpcUploadBinaryFileRequest createRequest() {
				return GrpcUploadBinaryFileRequest.newBuilder()
					.setMetadata(GrpcUploadBinaryFileMetaDataTestFactory.create())
					.build();
			}
		}

		@Nested
		class TestBuildChunkResponse {

			private static byte[] FILE_CONTENT = "file content".getBytes();

			@Test
			void shouldSetFileContent() {
				var grpcUploadBinaryFileRequest = service.buildChunkRequest(FILE_CONTENT, FILE_CONTENT.length);

				assertThat(grpcUploadBinaryFileRequest.getFileContent().toByteArray()).isEqualTo(FILE_CONTENT);
			}
		}

		@Nested
		class TestBuildCallStreamObserver {

			@Mock
			private StreamObserver<GrpcUploadBinaryFileResponse> responseObserver;
			@Mock
			private CallStreamObserver<GrpcUploadBinaryFileRequest> requestObserver;

			@BeforeEach
			void init() {
				doReturn(asyncServiceStub).when(service).getAsyncServiceStub();
			}

			@Test
			void shouldCallGetAsyncServiceStub() {
				service.buildCallStreamObserver(responseObserver);

				verify(service).getAsyncServiceStub();
			}

			@Test
			void shouldCallUploadBinaryFileAsStream() {
				 service.buildCallStreamObserver(responseObserver);

				verify(asyncServiceStub).uploadBinaryFileAsStream(responseObserver);
			}

			@Test
			void shouldReturnRequestObserver() {
				when(asyncServiceStub.uploadBinaryFileAsStream(any())).thenReturn(requestObserver);

				var result = service.buildCallStreamObserver(responseObserver);

				assertThat(result).isEqualTo(requestObserver);
			}
		}

		@Nested
		class TestBuildMetadataRequest {

			@Test
			void shouldSetMetadata() {
				var result = buildMetadataRequest();

				assertThat(result.getMetadata()).isNotEqualTo(GrpcUploadBinaryFileMetaData.getDefaultInstance());
			}

			@Test
			void shouldSetFileName() {
				var uploadFile = OzgCloudUploadFileTestFactory.create();

				var result = service.buildMetaDataRequest(uploadFile).getMetadata();

				assertThat(result.getFileName()).isEqualTo(uploadFile.getFileName());
			}

			@Test
			void shouldSetContentType() {
				var uploadFile = OzgCloudUploadFileTestFactory.create();

				var result = service.buildMetaDataRequest(uploadFile).getMetadata();

				assertThat(result.getContentType()).isEqualTo(uploadFile.getContentType());
			}

			@Test
			void shouldSetVorgangId() {
				var uploadFile = OzgCloudUploadFileTestFactory.create();

				var result = service.buildMetaDataRequest(uploadFile).getMetadata();

				assertThat(result.getVorgangId()).isEqualTo(uploadFile.getVorgangId().toString());
			}

			@Test
			void shouldSetField() {
				var uploadFile = OzgCloudUploadFileTestFactory.create();

				var result = service.buildMetaDataRequest(uploadFile).getMetadata();

				assertThat(result.getField()).isEqualTo(uploadFile.getFieldName());
			}

			private GrpcUploadBinaryFileRequest buildMetadataRequest() {
				return service.buildMetaDataRequest(OzgCloudUploadFileTestFactory.create());
			}
		}

		@Nested
		class TestWaitUntilFutureToComplete {

			private static final InputStream DATA_STREAM = InputStream.nullInputStream();

			@Mock
			private FileSender<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse> fileSender;

			@Mock
			private CompletableFuture<GrpcUploadBinaryFileResponse> future;

			@Test
			void shouldWaitUntilFutureToComplete() {
				when(fileSender.getResultFuture()).thenReturn(future);

				waitUntilFutureToComplete();

				verify(fileSender).getResultFuture();
			}

			@SneakyThrows
			@Test
			void shouldWaitWithTimeout() {
				when(fileSender.getResultFuture()).thenReturn(future);

				waitUntilFutureToComplete();

				verify(future).get(10, TimeUnit.MINUTES);
			}

			@Test
			void shouldCloseDatastream() {
				try(var ioUtils = Mockito.mockStatic(IOUtils.class)) {
					when(fileSender.getResultFuture()).thenReturn(future);

					waitUntilFutureToComplete();

					ioUtils.verify(() -> IOUtils.closeQuietly(DATA_STREAM));
				}
			}

			@Nested
			class TestThrowException {

				@BeforeEach
				void init() {
					when(fileSender.getResultFuture()).thenReturn(future);
				}

				@SneakyThrows
				@DisplayName("should throw TechnicalException")
				@ParameterizedTest(name = "when {0} is thrown")
				@ValueSource(classes = { InterruptedException.class, ExecutionException.class, TimeoutException.class })
				void shouldHandleInterruptedException(Class<? extends Exception> exceptionClass) {
					when(future.get(anyLong(), any())).thenThrow(exceptionClass);

					assertThrows(TechnicalException.class, TestWaitUntilFutureToComplete.this::waitUntilFutureToComplete);
				}

			}

			void waitUntilFutureToComplete() {
				service.waitUntilFutureToComplete(fileSender, DATA_STREAM);
			}
		}
	}
}
